# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gym.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="gym.proto",
    package="fetch.aea.Gym",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=b'\n\tgym.proto\x12\rfetch.aea.Gym"\xb1\x07\n\nGymMessage\x12\x12\n\nmessage_id\x18\x01 \x01(\x05\x12"\n\x1a\x64ialogue_starter_reference\x18\x02 \x01(\t\x12$\n\x1c\x64ialogue_responder_reference\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\x05\x12\x39\n\x03\x61\x63t\x18\x05 \x01(\x0b\x32*.fetch.aea.Gym.GymMessage.Act_PerformativeH\x00\x12=\n\x05\x63lose\x18\x06 \x01(\x0b\x32,.fetch.aea.Gym.GymMessage.Close_PerformativeH\x00\x12\x41\n\x07percept\x18\x07 \x01(\x0b\x32..fetch.aea.Gym.GymMessage.Percept_PerformativeH\x00\x12=\n\x05reset\x18\x08 \x01(\x0b\x32,.fetch.aea.Gym.GymMessage.Reset_PerformativeH\x00\x12?\n\x06status\x18\t \x01(\x0b\x32-.fetch.aea.Gym.GymMessage.Status_PerformativeH\x00\x1a\x18\n\tAnyObject\x12\x0b\n\x03\x61ny\x18\x01 \x01(\x0c\x1aX\n\x10\x41\x63t_Performative\x12\x33\n\x06\x61\x63tion\x18\x01 \x01(\x0b\x32#.fetch.aea.Gym.GymMessage.AnyObject\x12\x0f\n\x07step_id\x18\x02 \x01(\x05\x1a\xb2\x01\n\x14Percept_Performative\x12\x0f\n\x07step_id\x18\x01 \x01(\x05\x12\x38\n\x0bobservation\x18\x02 \x01(\x0b\x32#.fetch.aea.Gym.GymMessage.AnyObject\x12\x0e\n\x06reward\x18\x03 \x01(\x02\x12\x0c\n\x04\x64one\x18\x04 \x01(\x08\x12\x31\n\x04info\x18\x05 \x01(\x0b\x32#.fetch.aea.Gym.GymMessage.AnyObject\x1a\x92\x01\n\x13Status_Performative\x12K\n\x07\x63ontent\x18\x01 \x03(\x0b\x32:.fetch.aea.Gym.GymMessage.Status_Performative.ContentEntry\x1a.\n\x0c\x43ontentEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x14\n\x12Reset_Performative\x1a\x14\n\x12\x43lose_PerformativeB\x0e\n\x0cperformativeb\x06proto3',
)


_GYMMESSAGE_ANYOBJECT = _descriptor.Descriptor(
    name="AnyObject",
    full_name="fetch.aea.Gym.GymMessage.AnyObject",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="any",
            full_name="fetch.aea.Gym.GymMessage.AnyObject.any",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=470,
    serialized_end=494,
)

_GYMMESSAGE_ACT_PERFORMATIVE = _descriptor.Descriptor(
    name="Act_Performative",
    full_name="fetch.aea.Gym.GymMessage.Act_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="action",
            full_name="fetch.aea.Gym.GymMessage.Act_Performative.action",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="step_id",
            full_name="fetch.aea.Gym.GymMessage.Act_Performative.step_id",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=496,
    serialized_end=584,
)

_GYMMESSAGE_PERCEPT_PERFORMATIVE = _descriptor.Descriptor(
    name="Percept_Performative",
    full_name="fetch.aea.Gym.GymMessage.Percept_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="step_id",
            full_name="fetch.aea.Gym.GymMessage.Percept_Performative.step_id",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="observation",
            full_name="fetch.aea.Gym.GymMessage.Percept_Performative.observation",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="reward",
            full_name="fetch.aea.Gym.GymMessage.Percept_Performative.reward",
            index=2,
            number=3,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="done",
            full_name="fetch.aea.Gym.GymMessage.Percept_Performative.done",
            index=3,
            number=4,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="info",
            full_name="fetch.aea.Gym.GymMessage.Percept_Performative.info",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=587,
    serialized_end=765,
)

_GYMMESSAGE_STATUS_PERFORMATIVE_CONTENTENTRY = _descriptor.Descriptor(
    name="ContentEntry",
    full_name="fetch.aea.Gym.GymMessage.Status_Performative.ContentEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="fetch.aea.Gym.GymMessage.Status_Performative.ContentEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="fetch.aea.Gym.GymMessage.Status_Performative.ContentEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=868,
    serialized_end=914,
)

_GYMMESSAGE_STATUS_PERFORMATIVE = _descriptor.Descriptor(
    name="Status_Performative",
    full_name="fetch.aea.Gym.GymMessage.Status_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="content",
            full_name="fetch.aea.Gym.GymMessage.Status_Performative.content",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_GYMMESSAGE_STATUS_PERFORMATIVE_CONTENTENTRY,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=768,
    serialized_end=914,
)

_GYMMESSAGE_RESET_PERFORMATIVE = _descriptor.Descriptor(
    name="Reset_Performative",
    full_name="fetch.aea.Gym.GymMessage.Reset_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=916,
    serialized_end=936,
)

_GYMMESSAGE_CLOSE_PERFORMATIVE = _descriptor.Descriptor(
    name="Close_Performative",
    full_name="fetch.aea.Gym.GymMessage.Close_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=938,
    serialized_end=958,
)

_GYMMESSAGE = _descriptor.Descriptor(
    name="GymMessage",
    full_name="fetch.aea.Gym.GymMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="message_id",
            full_name="fetch.aea.Gym.GymMessage.message_id",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_starter_reference",
            full_name="fetch.aea.Gym.GymMessage.dialogue_starter_reference",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_responder_reference",
            full_name="fetch.aea.Gym.GymMessage.dialogue_responder_reference",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="target",
            full_name="fetch.aea.Gym.GymMessage.target",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="act",
            full_name="fetch.aea.Gym.GymMessage.act",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="close",
            full_name="fetch.aea.Gym.GymMessage.close",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="percept",
            full_name="fetch.aea.Gym.GymMessage.percept",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="reset",
            full_name="fetch.aea.Gym.GymMessage.reset",
            index=7,
            number=8,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="status",
            full_name="fetch.aea.Gym.GymMessage.status",
            index=8,
            number=9,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[
        _GYMMESSAGE_ANYOBJECT,
        _GYMMESSAGE_ACT_PERFORMATIVE,
        _GYMMESSAGE_PERCEPT_PERFORMATIVE,
        _GYMMESSAGE_STATUS_PERFORMATIVE,
        _GYMMESSAGE_RESET_PERFORMATIVE,
        _GYMMESSAGE_CLOSE_PERFORMATIVE,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="performative",
            full_name="fetch.aea.Gym.GymMessage.performative",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=29,
    serialized_end=974,
)

_GYMMESSAGE_ANYOBJECT.containing_type = _GYMMESSAGE
_GYMMESSAGE_ACT_PERFORMATIVE.fields_by_name[
    "action"
].message_type = _GYMMESSAGE_ANYOBJECT
_GYMMESSAGE_ACT_PERFORMATIVE.containing_type = _GYMMESSAGE
_GYMMESSAGE_PERCEPT_PERFORMATIVE.fields_by_name[
    "observation"
].message_type = _GYMMESSAGE_ANYOBJECT
_GYMMESSAGE_PERCEPT_PERFORMATIVE.fields_by_name[
    "info"
].message_type = _GYMMESSAGE_ANYOBJECT
_GYMMESSAGE_PERCEPT_PERFORMATIVE.containing_type = _GYMMESSAGE
_GYMMESSAGE_STATUS_PERFORMATIVE_CONTENTENTRY.containing_type = (
    _GYMMESSAGE_STATUS_PERFORMATIVE
)
_GYMMESSAGE_STATUS_PERFORMATIVE.fields_by_name[
    "content"
].message_type = _GYMMESSAGE_STATUS_PERFORMATIVE_CONTENTENTRY
_GYMMESSAGE_STATUS_PERFORMATIVE.containing_type = _GYMMESSAGE
_GYMMESSAGE_RESET_PERFORMATIVE.containing_type = _GYMMESSAGE
_GYMMESSAGE_CLOSE_PERFORMATIVE.containing_type = _GYMMESSAGE
_GYMMESSAGE.fields_by_name["act"].message_type = _GYMMESSAGE_ACT_PERFORMATIVE
_GYMMESSAGE.fields_by_name["close"].message_type = _GYMMESSAGE_CLOSE_PERFORMATIVE
_GYMMESSAGE.fields_by_name["percept"].message_type = _GYMMESSAGE_PERCEPT_PERFORMATIVE
_GYMMESSAGE.fields_by_name["reset"].message_type = _GYMMESSAGE_RESET_PERFORMATIVE
_GYMMESSAGE.fields_by_name["status"].message_type = _GYMMESSAGE_STATUS_PERFORMATIVE
_GYMMESSAGE.oneofs_by_name["performative"].fields.append(
    _GYMMESSAGE.fields_by_name["act"]
)
_GYMMESSAGE.fields_by_name["act"].containing_oneof = _GYMMESSAGE.oneofs_by_name[
    "performative"
]
_GYMMESSAGE.oneofs_by_name["performative"].fields.append(
    _GYMMESSAGE.fields_by_name["close"]
)
_GYMMESSAGE.fields_by_name["close"].containing_oneof = _GYMMESSAGE.oneofs_by_name[
    "performative"
]
_GYMMESSAGE.oneofs_by_name["performative"].fields.append(
    _GYMMESSAGE.fields_by_name["percept"]
)
_GYMMESSAGE.fields_by_name["percept"].containing_oneof = _GYMMESSAGE.oneofs_by_name[
    "performative"
]
_GYMMESSAGE.oneofs_by_name["performative"].fields.append(
    _GYMMESSAGE.fields_by_name["reset"]
)
_GYMMESSAGE.fields_by_name["reset"].containing_oneof = _GYMMESSAGE.oneofs_by_name[
    "performative"
]
_GYMMESSAGE.oneofs_by_name["performative"].fields.append(
    _GYMMESSAGE.fields_by_name["status"]
)
_GYMMESSAGE.fields_by_name["status"].containing_oneof = _GYMMESSAGE.oneofs_by_name[
    "performative"
]
DESCRIPTOR.message_types_by_name["GymMessage"] = _GYMMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GymMessage = _reflection.GeneratedProtocolMessageType(
    "GymMessage",
    (_message.Message,),
    {
        "AnyObject": _reflection.GeneratedProtocolMessageType(
            "AnyObject",
            (_message.Message,),
            {
                "DESCRIPTOR": _GYMMESSAGE_ANYOBJECT,
                "__module__": "gym_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Gym.GymMessage.AnyObject)
            },
        ),
        "Act_Performative": _reflection.GeneratedProtocolMessageType(
            "Act_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _GYMMESSAGE_ACT_PERFORMATIVE,
                "__module__": "gym_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Gym.GymMessage.Act_Performative)
            },
        ),
        "Percept_Performative": _reflection.GeneratedProtocolMessageType(
            "Percept_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _GYMMESSAGE_PERCEPT_PERFORMATIVE,
                "__module__": "gym_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Gym.GymMessage.Percept_Performative)
            },
        ),
        "Status_Performative": _reflection.GeneratedProtocolMessageType(
            "Status_Performative",
            (_message.Message,),
            {
                "ContentEntry": _reflection.GeneratedProtocolMessageType(
                    "ContentEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _GYMMESSAGE_STATUS_PERFORMATIVE_CONTENTENTRY,
                        "__module__": "gym_pb2"
                        # @@protoc_insertion_point(class_scope:fetch.aea.Gym.GymMessage.Status_Performative.ContentEntry)
                    },
                ),
                "DESCRIPTOR": _GYMMESSAGE_STATUS_PERFORMATIVE,
                "__module__": "gym_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Gym.GymMessage.Status_Performative)
            },
        ),
        "Reset_Performative": _reflection.GeneratedProtocolMessageType(
            "Reset_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _GYMMESSAGE_RESET_PERFORMATIVE,
                "__module__": "gym_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Gym.GymMessage.Reset_Performative)
            },
        ),
        "Close_Performative": _reflection.GeneratedProtocolMessageType(
            "Close_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _GYMMESSAGE_CLOSE_PERFORMATIVE,
                "__module__": "gym_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Gym.GymMessage.Close_Performative)
            },
        ),
        "DESCRIPTOR": _GYMMESSAGE,
        "__module__": "gym_pb2"
        # @@protoc_insertion_point(class_scope:fetch.aea.Gym.GymMessage)
    },
)
_sym_db.RegisterMessage(GymMessage)
_sym_db.RegisterMessage(GymMessage.AnyObject)
_sym_db.RegisterMessage(GymMessage.Act_Performative)
_sym_db.RegisterMessage(GymMessage.Percept_Performative)
_sym_db.RegisterMessage(GymMessage.Status_Performative)
_sym_db.RegisterMessage(GymMessage.Status_Performative.ContentEntry)
_sym_db.RegisterMessage(GymMessage.Reset_Performative)
_sym_db.RegisterMessage(GymMessage.Close_Performative)


_GYMMESSAGE_STATUS_PERFORMATIVE_CONTENTENTRY._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests of the messages module."""

import os
import shutil
import tempfile
from enum import Enum
from pathlib import Path
from typing import Callable, List, Tuple, Type

import pytest

from aea import AEA_DIR
from aea.configurations.constants import DEFAULT_PROTOCOL
from aea.exceptions import AEAEnforceError
from aea.mail.base import Envelope
from aea.protocols.base import Message, ProtobufSerializer, Protocol
from aea.protocols.default.dialogues import DefaultDialogue, DefaultDialogues
from aea.protocols.dialogue.base import Dialogue, DialogueLabel
from aea.protocols.signing.dialogues import SigningDialogue, SigningDialogues
from aea.protocols.state_update.dialogues import (
    StateUpdateDialogue,
    StateUpdateDialogues,
)

from tests.conftest import UNKNOWN_PROTOCOL_PUBLIC_ID


def role_from_first_message_dd(
    message: Message, receiver_address: str
) -> Dialogue.Role:
    """Role from first message."""
    return DefaultDialogue.Role.AGENT


def role_from_first_message_sd(
    message: Message, receiver_address: str
) -> Dialogue.Role:
    """Role from first message."""
    return SigningDialogue.Role.SKILL


def role_from_first_message_sud(
    message: Message, receiver_address: str
) -> Dialogue.Role:
    """Role from first message."""
    return StateUpdateDialogue.Role.SKILL


DIALOGUE_CLASSES: List[Tuple[Type, Type, Enum, Callable]] = [
    (
        DefaultDialogue,
        DefaultDialogues,
        DefaultDialogue.Role.AGENT,
        role_from_first_message_dd,
    ),
    (
        SigningDialogue,
        SigningDialogues,
        SigningDialogue.Role.SKILL,
        role_from_first_message_sd,
    ),
    (
        StateUpdateDialogue,
        StateUpdateDialogues,
        StateUpdateDialogue.Role.SKILL,
        role_from_first_message_sud,
    ),
]


class TestMessageProperties:
    """Test that the base serializations work."""

    @classmethod
    def setup_class(cls):
        """Setup test."""
        cls.body = {"body_1": "1", "body_2": "2"}
        cls.kwarg = 1
        cls.message = Message(cls.body, kwarg=cls.kwarg)

    def test_message_properties(self):
        """Test message properties."""
        for key, value in self.body.items():
            assert self.message.get(key) == value
        assert self.message.get("kwarg") == self.kwarg
        assert not self.message.has_sender
        assert not self.message.has_to
        to = "to"
        sender = "sender"
        self.message.to = to
        self.message.sender = sender
        assert self.message.sender == sender
        assert self.message.to == to
        assert (
            str(self.message)
            == "Message(sender=sender,to=to,body_1=1,body_2=2,kwarg=1)"
        )


class TestBaseSerializations:
    """Test that the base serializations work."""

    @classmethod
    def setup_class(cls):
        """Set up the use case."""
        cls.message = Message(content="hello")
        cls.message2 = Message(body={"content": "hello"})

    def test_default_protobuf_serialization(self):
        """Test that the default Protobuf serialization works."""
        message_bytes = ProtobufSerializer().encode(self.message)
        envelope = Envelope(
            to="receiver",
            sender="sender",
            protocol_id=UNKNOWN_PROTOCOL_PUBLIC_ID,
            message=message_bytes,
        )
        envelope_bytes = envelope.encode()

        expected_envelope = Envelope.decode(envelope_bytes)
        actual_envelope = envelope
        assert expected_envelope == actual_envelope

        expected_msg = ProtobufSerializer().decode(expected_envelope.message)
        actual_msg = self.message
        assert expected_msg == actual_msg

    def test_set(self):
        """Test that the set method works."""
        key, value = "temporary_key", "temporary_value"
        assert self.message.get(key) is None
        self.message.set(key, value)
        assert self.message.get(key) == value

    def test_body_setter(self):
        """Test the body setter."""
        m_dict = {"Hello": "World"}
        self.message2.body = m_dict
        assert "Hello" in self.message2.body.keys()


class TestProtocolFromDir:
    """Test the 'Protocol.from_dir' method."""

    @classmethod
    def setup_class(cls):
        """Set the tests up."""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

    def test_protocol_load_positive(self):
        """Test protocol loaded correctly."""
        default_protocol = Protocol.from_dir(Path(AEA_DIR, "protocols", "default"))
        assert str(default_protocol.public_id) == str(
            DEFAULT_PROTOCOL
        ), "Protocol not loaded correctly."
        assert default_protocol.serializer is not None

    @classmethod
    def teardown_class(cls):
        """Tear the tests down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestMessageAttributes:
    """Test some message attributes."""

    def test_performative(self):
        """Test message performative."""

        class SomePerformative(Message.Performative):
            value = "value"

        message = Message(performative=SomePerformative.value)
        assert message.performative == SomePerformative.value
        assert str(message.performative) == "value"

    def test_to(self):
        """Test the 'to' attribute getter and setter."""
        message = Message()
        with pytest.raises(ValueError, match="Message's 'To' field must be set."):
            message.to

        message.to = "to"
        assert message.to == "to"

        with pytest.raises(AEAEnforceError, match="To already set."):
            message.to = "to"

    def test_dialogue_reference(self):
        """Test the 'dialogue_reference' attribute."""
        message = Message(dialogue_reference=("x", "y"))
        assert message.dialogue_reference == ("x", "y")

    def test_message_id(self):
        """Test the 'message_id' attribute."""
        message = Message(message_id=1)
        assert message.message_id == 1

    def test_target(self):
        """Test the 'target' attribute."""
        message = Message(target=1)
        assert message.target == 1


@pytest.mark.parametrize("dialogue_classes", DIALOGUE_CLASSES)
def test_dialogue(dialogue_classes):
    """Test dialogue initialization."""
    dialogue_class, _, role, _ = dialogue_classes
    dialogue_class(
        DialogueLabel(("x", "y"), "opponent_addr", "starer_addr"), "agent_address", role
    )


@pytest.mark.parametrize("dialogues_classes", DIALOGUE_CLASSES)
def test_dialogues(dialogues_classes):
    """Test dialogues initialization."""
    dialogue_class, dialogues_class, _, role_from_first_message = dialogues_classes
    dialogues_class("agent_address", role_from_first_message, dialogue_class)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for aea/components/loader.py"""
from unittest import mock

import pytest

from aea.components.loader import load_component_from_config
from aea.configurations.base import ProtocolConfig
from aea.exceptions import AEAPackageLoadingError
from aea.protocols.base import Protocol


@pytest.fixture(scope="module")
def component_configuration():
    """Return a component configuration."""
    return ProtocolConfig("a_protocol", "an_author", "0.1.0")


def test_component_loading_generic_exception(component_configuration):
    """Test 'load_component_from_config' method when a generic "Exception" occurs."""

    with mock.patch.object(
        Protocol, "from_config", side_effect=Exception("Generic exception")
    ):
        with pytest.raises(
            Exception, match="An error occurred while loading .*: Generic exception"
        ):
            load_component_from_config(component_configuration)


def test_component_loading_generic_module_not_found_error(component_configuration):
    """Test 'load_component_from_config' method when a generic "ModuleNotFoundError" occurs."""

    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError(
            "An error occurred while loading .*: Generic error"
        ),
    ):
        with pytest.raises(ModuleNotFoundError, match="Generic error"):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_non_framework_package(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a generic import path (non framework related."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError("No module named 'generic.package'"),
    ):
        with pytest.raises(ModuleNotFoundError):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_framework_package(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') but for some reason it doesn't contain the author name."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError("No module named 'packages'"),
    ):
        with pytest.raises(ModuleNotFoundError, match="No module named 'packages'"):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_framework_package_with_wrong_author(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') with wrong author."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError("No module named 'packages.some_author'"),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match="No module named packages.some_author; No AEA package found with author name 'some_author'",
        ):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_framework_package_with_wrong_type(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') with correct author but wrong type."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError(
            "No module named 'packages.some_author.some_type'"
        ),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match=r"No module named packages.some_author.some_type; 'some_type' is not a valid type name, choose one of \['protocols', 'connections', 'skills', 'contracts'\]",
        ):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_framework_package_with_wrong_name(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') with correct author and type but wrong name."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError(
            "No module named 'packages.some_author.protocols.some_name'"
        ),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match="No module named packages.some_author.protocols.some_name; No AEA package found with author name 'some_author', type 'protocols', name 'some_name'",
        ):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_framework_package_with_wrong_suffix(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') with correct author and type but wrong suffix."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError(
            "No module named 'packages.some_author.protocols.some_name.some_subpackage'"
        ),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match="No module named packages.some_author.protocols.some_name.some_subpackage; The package 'packages/some_author' of type 'protocols' exists, but cannot find module 'some_subpackage'",
        ):
            load_component_from_config(component_configuration)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the dialogue/base.py module."""

import sys
from typing import FrozenSet, Tuple, Type, cast
from unittest import mock

import pytest

import aea
from aea.common import Address
from aea.exceptions import AEAEnforceError
from aea.protocols.base import Message
from aea.protocols.default.message import DefaultMessage
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel, DialogueMessage, DialogueStats
from aea.protocols.dialogue.base import Dialogues as BaseDialogues
from aea.protocols.dialogue.base import InvalidDialogueMessage
from aea.protocols.state_update.message import StateUpdateMessage


class Dialogue(BaseDialogue):
    """This concrete class defines a dialogue."""

    INITIAL_PERFORMATIVES = frozenset({DefaultMessage.Performative.BYTES})
    TERMINAL_PERFORMATIVES = frozenset({DefaultMessage.Performative.ERROR})
    VALID_REPLIES = {
        DefaultMessage.Performative.BYTES: frozenset(
            {DefaultMessage.Performative.BYTES, DefaultMessage.Performative.ERROR}
        ),
        DefaultMessage.Performative.ERROR: frozenset(),
    }

    class Role(BaseDialogue.Role):
        """This class defines the agent's role in this dialogue."""

        ROLE1 = "role1"
        ROLE2 = "role2"

    class EndState(BaseDialogue.EndState):
        """This class defines the end states of this dialogue."""

        SUCCESSFUL = 0
        FAILED = 1

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        message_class: Type[Message] = DefaultMessage,
        self_address: Address = "agent 1",
        role: BaseDialogue.Role = Role.ROLE1,
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :return: None
        """
        BaseDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class Dialogues(BaseDialogues):
    """This class gives a concrete definition of dialogues."""

    END_STATES = frozenset(
        {Dialogue.EndState.SUCCESSFUL, Dialogue.EndState.FAILED}
    )  # type: FrozenSet[BaseDialogue.EndState]

    def __init__(
        self,
        self_address: Address,
        message_class=DefaultMessage,
        dialogue_class=Dialogue,
    ) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom dialogues are maintained
        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return Dialogue.Role.ROLE1

        BaseDialogues.__init__(
            self,
            self_address=self_address,
            end_states=cast(FrozenSet[BaseDialogue.EndState], self.END_STATES),
            message_class=message_class,
            dialogue_class=dialogue_class,
            role_from_first_message=role_from_first_message,
        )


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="This part of code is only defined for python version >= 3.7",
)
def test_dialogue_message_python_3_7():
    """Test DiallogueMessage if python is 3.7"""
    dialogue_message = DialogueMessage(DefaultMessage.Performative.BYTES)
    assert isinstance(dialogue_message.performative, Message.Performative)

    assert dialogue_message.performative == DefaultMessage.Performative.BYTES
    assert dialogue_message.contents == {}
    assert dialogue_message.is_incoming is None
    assert dialogue_message.target is None


@pytest.mark.skipif(
    sys.version_info >= (3, 7),
    reason="This part of code is only defined for python version < 3.7",
)
def test_dialogue_message_python_3_6():
    """Test DiallogueMessage if python is 3.6"""
    with mock.patch.object(
        aea.protocols.dialogue.base.sys, "version_info", return_value=(3, 6)
    ):
        dialogue_message = DialogueMessage(DefaultMessage.Performative.BYTES)
        assert isinstance(dialogue_message.performative, Message.Performative)

        assert dialogue_message.performative == DefaultMessage.Performative.BYTES
        assert dialogue_message.contents == {}
        assert dialogue_message.is_incoming is None
        assert dialogue_message.target is None


class TestDialogueLabel:
    """Test for DialogueLabel."""

    @classmethod
    def setup(cls):
        """Initialise the environment to test DialogueLabel."""
        cls.agent_address = "agent 1"
        cls.opponent_address = "agent 2"
        cls.dialogue_label = DialogueLabel(
            dialogue_reference=(str(1), ""),
            dialogue_opponent_addr=cls.opponent_address,
            dialogue_starter_addr=cls.agent_address,
        )

    def test_all_methods(self):
        """Test the DialogueLabel."""
        assert self.dialogue_label.dialogue_reference == (str(1), "")
        assert self.dialogue_label.dialogue_starter_reference == str(1)
        assert self.dialogue_label.dialogue_responder_reference == ""
        assert self.dialogue_label.dialogue_opponent_addr == self.opponent_address
        assert self.dialogue_label.dialogue_starter_addr == self.agent_address
        assert str(self.dialogue_label) == "{}_{}_{}_{}".format(
            self.dialogue_label.dialogue_starter_reference,
            self.dialogue_label.dialogue_responder_reference,
            self.dialogue_label.dialogue_opponent_addr,
            self.dialogue_label.dialogue_starter_addr,
        )

        dialogue_label_eq = DialogueLabel(
            dialogue_reference=(str(1), ""),
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )

        assert dialogue_label_eq == self.dialogue_label

        dialogue_label_not_eq = "This is a test"

        assert not dialogue_label_not_eq == self.dialogue_label

        assert hash(dialogue_label_eq) == hash(self.dialogue_label)

        assert self.dialogue_label.json == dict(
            dialogue_starter_reference=str(1),
            dialogue_responder_reference="",
            dialogue_opponent_addr=self.opponent_address,
            dialogue_starter_addr=self.agent_address,
        )
        assert DialogueLabel.from_json(self.dialogue_label.json) == self.dialogue_label
        assert DialogueLabel.from_str(str(self.dialogue_label)) == self.dialogue_label


class TestDialogueBase:
    """Test for Dialogue."""

    @classmethod
    def setup(cls):
        """Initialise the environment to test Dialogue."""
        cls.incomplete_reference = (str(1), "")
        cls.complete_reference = (str(1), str(1))
        cls.opponent_address = "agent 2"
        cls.agent_address = "agent 1"

        cls.dialogue_label = DialogueLabel(
            dialogue_reference=cls.incomplete_reference,
            dialogue_opponent_addr=cls.opponent_address,
            dialogue_starter_addr=cls.agent_address,
        )
        cls.dialogue = Dialogue(dialogue_label=cls.dialogue_label)

        cls.dialogue_label_opponent_started = DialogueLabel(
            dialogue_reference=cls.complete_reference,
            dialogue_opponent_addr=cls.opponent_address,
            dialogue_starter_addr=cls.opponent_address,
        )
        cls.dialogue_opponent_started = Dialogue(
            dialogue_label=cls.dialogue_label_opponent_started
        )

        # convenient messages to reuse across tests
        cls.valid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        cls.valid_message_1_by_self.sender = cls.agent_address
        cls.valid_message_1_by_self.to = cls.opponent_address

        cls.valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        cls.valid_message_2_by_other.sender = cls.opponent_address
        cls.valid_message_2_by_other.to = cls.agent_address

        cls.valid_message_3_by_self = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=3,
            target=2,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back 2",
        )
        cls.valid_message_3_by_self.sender = cls.agent_address
        cls.valid_message_3_by_self.to = cls.opponent_address

    def test_inner_classes(self):
        """Test the inner classes: Role and EndStates."""
        assert str(Dialogue.Role.ROLE1) == "role1"
        assert str(Dialogue.Role.ROLE2) == "role2"
        assert str(Dialogue.EndState.SUCCESSFUL) == "0"
        assert str(Dialogue.EndState.FAILED) == "1"

    def test_dialogue_properties(self):
        """Test dialogue properties."""
        assert self.dialogue.dialogue_label == self.dialogue_label
        assert self.dialogue.incomplete_dialogue_label == self.dialogue_label
        assert self.dialogue.dialogue_labels == {self.dialogue_label}
        assert self.dialogue.self_address == self.agent_address

        assert self.dialogue.role == Dialogue.Role.ROLE1
        assert str(self.dialogue.role) == "role1"

        assert self.dialogue.rules.initial_performatives == frozenset(
            {DefaultMessage.Performative.BYTES}
        )
        assert self.dialogue.rules.terminal_performatives == frozenset(
            {DefaultMessage.Performative.ERROR}
        )
        assert self.dialogue.rules.valid_replies == {
            DefaultMessage.Performative.BYTES: frozenset(
                {DefaultMessage.Performative.BYTES, DefaultMessage.Performative.ERROR}
            ),
            DefaultMessage.Performative.ERROR: frozenset(),
        }
        assert self.dialogue.rules.get_valid_replies(
            DefaultMessage.Performative.BYTES
        ) == frozenset(
            {DefaultMessage.Performative.BYTES, DefaultMessage.Performative.ERROR}
        )
        assert self.dialogue.rules.get_valid_replies(
            DefaultMessage.Performative.ERROR
        ) == frozenset({})
        assert self.dialogue.message_class == DefaultMessage

        assert self.dialogue.is_self_initiated

        assert self.dialogue.last_incoming_message is None
        assert self.dialogue.last_outgoing_message is None
        assert self.dialogue.last_message is None

        assert self.dialogue.is_empty

    def test_counterparty_from_message(self):
        """Test the 'counterparty_from_message' method."""
        assert (
            self.dialogue._counterparty_from_message(self.valid_message_1_by_self)
            == self.opponent_address
        )
        assert (
            self.dialogue._counterparty_from_message(self.valid_message_2_by_other)
            == self.opponent_address
        )

    def test_is_message_by_self(self):
        """Test the 'is_message_by_self' method."""
        assert self.dialogue._is_message_by_self(self.valid_message_1_by_self)
        assert not self.dialogue._is_message_by_self(self.valid_message_2_by_other)

    def test_is_message_by_other(self):
        """Test the 'is_message_by_other' method."""
        assert not self.dialogue._is_message_by_other(self.valid_message_1_by_self)
        assert self.dialogue._is_message_by_other(self.valid_message_2_by_other)

    def test_try_get_message(self):
        """Test the 'try_get_message' method."""
        assert (
            self.dialogue._try_get_message(self.valid_message_1_by_self.message_id)
            is None
        )
        self.dialogue._update(self.valid_message_1_by_self)
        assert (
            self.dialogue._try_get_message(self.valid_message_1_by_self.message_id)
            == self.valid_message_1_by_self
        )

        assert (
            self.dialogue._try_get_message(self.valid_message_2_by_other.message_id)
            is None
        )
        self.dialogue._update(self.valid_message_2_by_other)
        assert (
            self.dialogue._try_get_message(self.valid_message_2_by_other.message_id)
            == self.valid_message_2_by_other
        )

    def test_get_message(self):
        """Test the 'get_message' method."""
        with pytest.raises(ValueError) as cm:
            self.dialogue._get_message(self.valid_message_1_by_self.message_id)
        assert str(cm.value) == "Message not present."

        self.dialogue._update(self.valid_message_1_by_self)
        assert (
            self.dialogue._get_message(self.valid_message_1_by_self.message_id)
            == self.valid_message_1_by_self
        )

        with pytest.raises(ValueError) as cm:
            self.dialogue._get_message(self.valid_message_2_by_other.message_id)
        assert str(cm.value) == "Message not present."

        self.dialogue._update(self.valid_message_2_by_other)
        assert (
            self.dialogue._get_message(self.valid_message_2_by_other.message_id)
            == self.valid_message_2_by_other
        )

    def test_has_message(self):
        """Test the 'has_message' method."""
        assert self.dialogue._has_message(self.valid_message_1_by_self) is False

        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue._has_message(self.valid_message_1_by_self) is True

        assert self.dialogue._has_message(self.valid_message_2_by_other) is False

    def test_has_message_id(self):
        """Test the 'has_message_id' method."""
        assert self.dialogue._has_message_id(1) is False

        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue._has_message_id(1) is True

        assert self.dialogue._has_message_id(2) is False

    def test_update_positive(self):
        """Positive test for the 'update' method."""
        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue.last_outgoing_message == self.valid_message_1_by_self

    def test_update_positive_multiple_messages_by_self(self):
        """Positive test for the 'update' method: multiple messages by self is sent to the dialogue."""
        self.dialogue._update(self.valid_message_1_by_self)

        valid_message_2_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_self.sender = self.agent_address
        valid_message_2_by_self.to = self.opponent_address

        self.dialogue._update(valid_message_2_by_self)

        assert self.dialogue.last_message.message_id == 2

    def test_update_negative_is_valid_next_message_fails(self):
        """Negative test for the 'update' method: input message is invalid with respect to the dialogue."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        with pytest.raises(InvalidDialogueMessage) as cm:
            self.dialogue._update(invalid_message_1_by_self)
        assert (
            str(cm.value)
            == "Message 2 is invalid with respect to this dialogue. Error: "
            + "Invalid message_id. Expected 1. Found 2."
        )

        assert self.dialogue.last_outgoing_message is None

    def test_update_dialogue_negative_message_does_not_belong_to_dialogue(self):
        """Negative test for the 'update' method in dialogue with wrong message not belonging to dialogue."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(2), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        with pytest.raises(InvalidDialogueMessage) as cm:
            self.dialogue._update(invalid_message_1_by_self)
        assert str(cm.value) == (
            "The message 1 does not belong to this dialogue."
            "The dialogue reference of the message is {}, while the dialogue reference of the dialogue is {}".format(
                invalid_message_1_by_self.dialogue_reference,
                self.dialogue.dialogue_label.dialogue_reference,
            )
        )
        assert self.dialogue.is_empty

    def test_is_belonging_to_dialogue(self):
        """Test for the '_is_belonging_to_dialogue' method"""
        valid_message_2_by_self = DefaultMessage(
            dialogue_reference=(str(2), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        valid_message_2_by_self.sender = self.agent_address
        valid_message_2_by_self.to = self.opponent_address

        assert self.dialogue._is_belonging_to_dialogue(self.valid_message_1_by_self)
        assert not self.dialogue._is_belonging_to_dialogue(valid_message_2_by_self)

    def test_reply_positive(self):
        """Positive test for the 'reply' method."""
        self.dialogue._update(self.valid_message_1_by_self)

        self.dialogue.reply(
            target_message=self.valid_message_1_by_self,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello Back",
        )

        assert self.dialogue.last_message.message_id == 2

    def test_reply_negative_empty_dialogue(self):
        """Negative test for the 'reply' method: target message is not in the dialogue."""
        with pytest.raises(ValueError) as cm:
            self.dialogue.reply(
                target_message=self.valid_message_1_by_self,
                performative=DefaultMessage.Performative.BYTES,
                content=b"Hello Back",
            )
        assert str(cm.value) == "Cannot reply in an empty dialogue!"
        assert self.dialogue.is_empty

    def test_reply_negative_target_message_target_mismatch(self):
        """Negative test for the 'reply' method: target message and target provided but do not match."""
        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue.last_message.message_id == 1

        with pytest.raises(AEAEnforceError) as cm:
            self.dialogue.reply(
                target_message=self.valid_message_1_by_self,
                target=2,
                performative=DefaultMessage.Performative.BYTES,
                content=b"Hello Back",
            )
        assert str(cm.value) == "The provided target and target_message do not match."
        assert self.dialogue.last_message.message_id == 1

    def test_reply_negative_invalid_target(self):
        """Negative test for the 'reply' method: target message is not in the dialogue."""
        self.dialogue._update(self.valid_message_1_by_self)
        assert self.dialogue.last_message.message_id == 1

        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello There",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        with pytest.raises(AEAEnforceError) as cm:
            self.dialogue.reply(
                target_message=invalid_message_1_by_self,
                performative=DefaultMessage.Performative.BYTES,
                content=b"Hello Back",
            )
        assert str(cm.value) == "The target message does not exist in this dialogue."
        assert self.dialogue.last_message.message_id == 1

    def test_is_valid_next_message_positive(self):
        """Positive test for the 'validate_next_message' method"""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        result, msg = self.dialogue._validate_next_message(self.valid_message_3_by_self)
        assert result is True
        assert msg == "Message is valid with respect to this dialogue."

    def test_is_valid_next_message_negative_basic_validation_fails(self):
        """Negative test for the 'validate_next_message' method: basic_validation method fails"""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._validate_next_message(invalid_message_1_by_self)
        assert result is False
        assert msg == "Invalid message_id. Expected 1. Found 2."

    def test_is_valid_next_message_negative_additional_validation_fails(self):
        """Negative test for the 'validate_next_message' method: additional_validation method fails"""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        invalid_message_3_by_self = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=3,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back 2",
        )
        invalid_message_3_by_self.sender = self.agent_address
        invalid_message_3_by_self.to = self.opponent_address

        result, msg = self.dialogue._validate_next_message(invalid_message_3_by_self)
        assert result is False
        assert msg == "Invalid target. Expected 2. Found 1."

    def test_is_valid_next_message_negative_is_valid_fails(self):
        """Negative test for the 'validate_next_message' method: is_valid method fails"""

        def failing_custom_validation(message: Message) -> Tuple[bool, str]:
            return False, "some reason"

        self.dialogue._custom_validation = failing_custom_validation

        result, msg = self.dialogue._validate_next_message(self.valid_message_1_by_self)
        assert result is False
        assert msg == "some reason"

    def test_basic_validation_positive(self):
        """Positive test for the '_basic_validation' method."""
        result, msg = self.dialogue._basic_validation(self.valid_message_1_by_self)
        assert result is True
        assert msg == "The initial message passes basic validation."

        self.dialogue._update(self.valid_message_1_by_self)

        result, msg = self.dialogue._basic_validation(self.valid_message_2_by_other)
        assert result is True
        assert msg == "The non-initial message passes basic validation."

    def test_basic_validation_negative_initial_message_invalid(self):
        """Negative test for the '_basic_validation' method: initial message is invalid."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._basic_validation(invalid_message_1_by_self)
        assert result is False
        assert msg == "Invalid message_id. Expected 1. Found 2."

    def test_basic_validation_negative_non_initial_message_invalid(self):
        """Negative test for the '_basic_validation' method: non-initial message is invalid."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation(invalid_message_2_by_other)
        assert result is False
        assert (
            msg
            == "Invalid target. Expected a value greater than or equal to 1. Found 0."
        )

    def test_basic_validation_initial_message_positive(self):
        """Positive test for the '_basic_validation_initial_message' method."""
        result, msg = self.dialogue._basic_validation_initial_message(
            self.valid_message_1_by_self
        )
        assert result is True
        assert msg == "The initial message passes basic validation."

    def test_basic_validation_initial_message_negative_invalid_dialogue_reference(self):
        """Negative test for the '_basic_validation' method: input message has invalid dialogue reference."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(2), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._basic_validation_initial_message(
            invalid_message_1_by_self
        )
        assert result is False
        assert msg == "Invalid dialogue_reference[0]. Expected 1. Found 2."

    def test_basic_validation_initial_message_negative_invalid_message_id(self):
        """Negative test for the '_basic_validation' method: input message has invalid message id."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._basic_validation_initial_message(
            invalid_message_1_by_self
        )
        assert result is False
        assert msg == "Invalid message_id. Expected 1. Found 2."

    def test_basic_validation_initial_message_negative_invalid_target(self):
        """Negative test for the '_basic_validation_initial_message' method: input message has invalid target."""
        invalid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            message_id=1,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_self.sender = self.agent_address
        invalid_message_1_by_self.to = self.opponent_address

        result, msg = self.dialogue._basic_validation_initial_message(
            invalid_message_1_by_self
        )
        assert result is False
        assert msg == "Invalid target. Expected 0. Found 1."

    def test_basic_validation_initial_message_negative_invalid_performative(self):
        """Negative test for the '_basic_validation_initial_message' method: input message has invalid performative."""
        invalid_initial_msg = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.ERROR,
            error_code=DefaultMessage.ErrorCode.INVALID_MESSAGE,
            error_msg="some_error_message",
            error_data={"some_data": b"some_bytes"},
        )
        invalid_initial_msg.sender = self.agent_address
        invalid_initial_msg.to = self.opponent_address

        result, msg = self.dialogue._basic_validation_initial_message(
            invalid_initial_msg
        )
        assert result is False
        assert (
            msg
            == "Invalid initial performative. Expected one of {}. Found error.".format(
                self.dialogue.rules.initial_performatives
            )
        )

    def test_basic_validation_non_initial_message_positive(self):
        """Positive test for the '_basic_validation_non_initial_message' method."""
        self.dialogue._update(self.valid_message_1_by_self)

        result, msg = self.dialogue._basic_validation_non_initial_message(
            self.valid_message_2_by_other
        )
        assert result is True
        assert msg == "The non-initial message passes basic validation."

    def test_basic_validation_non_initial_message_negative_invalid_dialogue_reference(
        self,
    ):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has invalid dialogue reference."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(2), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert msg == "Invalid dialogue_reference[0]. Expected 1. Found 2."

    def test_basic_validation_non_initial_message_negative_invalid_message_id(self):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has invalid message id."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=3,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert msg == "Invalid message_id. Expected 2. Found 3."

    def test_basic_validation_non_initial_message_negative_invalid_target_1(self):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has target less than 1."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert (
            msg
            == "Invalid target. Expected a value greater than or equal to 1. Found 0."
        )

    def test_basic_validation_non_initial_message_negative_invalid_target_2(self):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has target greater than the id of the last existing message."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=2,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert (
            msg == "Invalid target. Expected a value less than or equal to 1. Found 2."
        )

    def test_basic_validation_non_initial_message_negative_invalid_performative(self):
        """Negative test for the '_basic_validation_non_initial_message' method: input message has invalid performative."""
        self.dialogue._update(self.valid_message_1_by_self)

        invalid_message_2_by_other = StateUpdateMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=1,
            performative=StateUpdateMessage.Performative.APPLY,
            amount_by_currency_id={},
            quantities_by_good_id={},
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        result, msg = self.dialogue._basic_validation_non_initial_message(
            invalid_message_2_by_other
        )
        assert result is False
        assert msg == "Invalid performative. Expected one of {}. Found {}.".format(
            self.dialogue.rules.get_valid_replies(
                self.valid_message_1_by_self.performative
            ),
            invalid_message_2_by_other.performative,
        )

    def test_additional_validation_positive(self):
        """Positive test for the '_additional_rules' method."""
        result, msg = self.dialogue._additional_validation(self.valid_message_1_by_self)
        assert result is True
        assert msg == "The message passes additional validation."

        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        result, msg = self.dialogue._additional_validation(self.valid_message_3_by_self)
        assert result is True
        assert msg == "The message passes additional validation."

    def test_additional_validation_negative_invalid_target(self):
        """Negative test for the '_additional_rules' method: input message has invalid target (its target is not the last message of the dialogue)."""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        invalid_message_3_by_self = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=3,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back 2",
        )
        invalid_message_3_by_self.sender = self.agent_address
        invalid_message_3_by_self.to = self.opponent_address

        result, msg = self.dialogue._additional_validation(invalid_message_3_by_self)
        assert result is False
        assert msg == "Invalid target. Expected 2. Found 1."

    def test_update_dialogue_label_positive(self):
        """Positive test for the 'update_dialogue_label' method."""
        self.dialogue._update(self.valid_message_1_by_self)

        new_label = DialogueLabel(
            (str(1), str(1)), self.valid_message_1_by_self.to, self.agent_address
        )
        self.dialogue._update_dialogue_label(new_label)

        assert self.dialogue.dialogue_label == new_label

    def test_update_dialogue_label_negative_invalid_existing_label(self):
        """Negative test for the 'update_dialogue_label' method: existing dialogue reference is invalid."""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        new_label = DialogueLabel(
            (str(1), str(1)), self.valid_message_1_by_self.to, self.agent_address
        )
        self.dialogue._update_dialogue_label(new_label)
        assert self.dialogue.dialogue_label == new_label

        new_label = DialogueLabel(
            (str(1), str(2)), self.valid_message_1_by_self.to, self.agent_address
        )
        with pytest.raises(AEAEnforceError) as cm:
            self.dialogue._update_dialogue_label(new_label)
        assert str(cm.value) == "Dialogue label cannot be updated."

        assert self.dialogue.dialogue_label != new_label

    def test_update_dialogue_label_negative_invalid_input_label(self):
        """Negative test for the 'update_dialogue_label' method: input dialogue label's dialogue reference is invalid."""
        self.dialogue._update(self.valid_message_1_by_self)

        new_label = DialogueLabel(
            (str(2), ""), self.valid_message_1_by_self.to, self.agent_address
        )
        with pytest.raises(AEAEnforceError) as cm:
            self.dialogue._update_dialogue_label(new_label)
        assert str(cm.value) == "Dialogue label cannot be updated."
        assert self.dialogue.dialogue_label != new_label

    def test_interleave(self):
        """Test the '_interleave' method."""
        list_1 = [1, 3, 5, 7]
        list_2 = [2, 4, 6, 8]
        assert Dialogue._interleave(list_1, list_2) == [1, 2, 3, 4, 5, 6, 7, 8]

        list_3 = [1, 3, 4]
        list_4 = [2]
        assert Dialogue._interleave(list_3, list_4) == [1, 2, 3, 4]

    def test___str__1(self):
        """Test the '__str__' method: dialogue is self initiated"""
        self.dialogue._update(self.valid_message_1_by_self)
        self.dialogue._update(self.valid_message_2_by_other)

        self.dialogue._update(self.valid_message_3_by_self)

        dialogue_str = (
            "Dialogue Label: 1__agent 2_agent 1\nbytes( )\nbytes( )\nbytes( )"
        )

        assert str(self.dialogue) == dialogue_str

    def test___str__2(self):
        """Test the '__str__' method: dialogue is other initiated"""
        valid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        valid_message_1_by_other.sender = self.opponent_address
        valid_message_1_by_other.to = self.agent_address

        self.dialogue_opponent_started._update(valid_message_1_by_other)

        valid_message_2_by_self = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_self.sender = self.agent_address
        valid_message_2_by_self.to = self.opponent_address

        self.dialogue_opponent_started._update(valid_message_2_by_self)

        valid_message_3_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=3,
            target=2,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back 2",
        )
        valid_message_3_by_other.sender = self.opponent_address
        valid_message_3_by_other.to = self.agent_address

        self.dialogue_opponent_started._update(valid_message_3_by_other)

        dialogue_str = (
            "Dialogue Label: 1_1_agent 2_agent 2\nbytes( )\nbytes( )\nbytes( )"
        )

        assert str(self.dialogue_opponent_started) == dialogue_str


class TestDialogueStats:
    """Test for DialogueStats."""

    @classmethod
    def setup(cls):
        """Initialise the environment to test DialogueStats."""
        cls.agent_address = "agent 1"
        cls.opponent_address = "agent 2"
        cls.dialogue_label = DialogueLabel(
            dialogue_reference=(str(1), ""),
            dialogue_opponent_addr=cls.opponent_address,
            dialogue_starter_addr=cls.agent_address,
        )
        cls.dialogue = Dialogue(dialogue_label=cls.dialogue_label)
        end_states = frozenset(
            {Dialogue.EndState.SUCCESSFUL, Dialogue.EndState.FAILED}
        )  # type: FrozenSet[BaseDialogue.EndState]
        cls.dialogue_stats = DialogueStats(end_states)

    def test_properties(self):
        """Test dialogue properties."""
        assert isinstance(self.dialogue_stats.self_initiated, dict)
        assert self.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }

    def test_add_dialogue_endstate(self):
        """Test for the 'add_dialogue_endstate' method."""
        assert self.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }

        self.dialogue_stats.add_dialogue_endstate(Dialogue.EndState.SUCCESSFUL, True)
        assert self.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 1,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }

        self.dialogue_stats.add_dialogue_endstate(Dialogue.EndState.FAILED, False)
        assert self.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 1,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 1,
        }


class TestDialoguesBase:
    """Test for Dialogues."""

    @classmethod
    def setup(cls):
        """Initialise the environment to test Dialogue."""
        cls.agent_address = "agent 1"
        cls.opponent_address = "agent 2"
        cls.dialogue_label = DialogueLabel(
            dialogue_reference=(str(1), ""),
            dialogue_opponent_addr=cls.opponent_address,
            dialogue_starter_addr=cls.agent_address,
        )
        cls.dialogue = Dialogue(dialogue_label=cls.dialogue_label)
        cls.own_dialogues = Dialogues(cls.agent_address)
        cls.opponent_dialogues = Dialogues(cls.opponent_address)

        # convenient messages to reuse across tests
        cls.valid_message_1_by_self = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        cls.valid_message_1_by_self.sender = cls.agent_address
        cls.valid_message_1_by_self.to = cls.opponent_address

        cls.valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(1), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        cls.valid_message_2_by_other.sender = cls.opponent_address
        cls.valid_message_2_by_other.to = cls.agent_address

    def test_dialogues_properties(self):
        """Test dialogue properties."""
        assert self.own_dialogues.dialogues == dict()
        assert self.own_dialogues.self_address == self.agent_address
        assert self.own_dialogues.dialogue_stats.other_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.own_dialogues.dialogue_stats.self_initiated == {
            Dialogue.EndState.SUCCESSFUL: 0,
            Dialogue.EndState.FAILED: 0,
        }
        assert self.own_dialogues.message_class == DefaultMessage
        assert self.own_dialogues.dialogue_class == Dialogue

    def test_counterparty_from_message(self):
        """Test the 'counterparty_from_message' method."""
        assert (
            self.own_dialogues._counterparty_from_message(self.valid_message_1_by_self)
            == self.opponent_address
        )
        assert (
            self.own_dialogues._counterparty_from_message(self.valid_message_2_by_other)
            == self.opponent_address
        )

    def test_is_message_by_self(self):
        """Test the 'is_message_by_self' method."""
        assert self.own_dialogues._is_message_by_self(self.valid_message_1_by_self)
        assert not self.own_dialogues._is_message_by_self(self.valid_message_2_by_other)

    def test_is_message_by_other(self):
        """Test the 'is_message_by_other' method."""
        assert not self.own_dialogues._is_message_by_other(self.valid_message_1_by_self)
        assert self.own_dialogues._is_message_by_other(self.valid_message_2_by_other)

    def test_new_self_initiated_dialogue_reference(self):
        """Test the 'new_self_initiated_dialogue_reference' method."""
        self_initiated_ref = self.own_dialogues.new_self_initiated_dialogue_reference()
        assert (
            isinstance(self_initiated_ref[0], str)
            and self_initiated_ref[0] != ""
            and len(self_initiated_ref[0]) == DialogueLabel.NONCE_BYTES_NB * 2
        )
        assert self_initiated_ref[1] == ""
        self_initiated_ref_2 = (
            self.own_dialogues.new_self_initiated_dialogue_reference()
        )
        assert self_initiated_ref_2 != self_initiated_ref

    def test_create_positive(self):
        """Positive test for the 'create' method."""
        assert len(self.own_dialogues.dialogues) == 0
        self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )
        assert len(self.own_dialogues.dialogues) == 1

    def test_create_negative_incorrect_performative_content_combination(self):
        """Negative test for the 'create' method: invalid performative and content combination (i.e. invalid message)."""
        assert len(self.own_dialogues.dialogues) == 0
        with pytest.raises(SyntaxError) as cm:
            self.own_dialogues.create(
                self.opponent_address,
                DefaultMessage.Performative.ERROR,
                content=b"Hello",
            )
        assert (
            str(cm.value)
            == "Cannot create a dialogue with the specified performative and contents."
        )
        assert len(self.own_dialogues.dialogues) == 0

    def test_update_positive_new_dialogue_by_other(self):
        """Positive test for the 'update' method: the input message is for a new dialogue dialogue by other."""
        valid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        valid_message_1_by_other.sender = self.opponent_address
        valid_message_1_by_other.to = self.agent_address

        assert len(self.own_dialogues.dialogues) == 0

        dialogue = self.own_dialogues.update(valid_message_1_by_other)

        assert len(self.own_dialogues.dialogues) == 1
        assert dialogue is not None
        assert dialogue.last_message.dialogue_reference == (str(1), "")
        assert dialogue.last_message.message_id == 1
        assert dialogue.last_message.target == 0
        assert dialogue.last_message.performative == DefaultMessage.Performative.BYTES
        assert dialogue.last_message.content == b"Hello"

    def test_update_positive_existing_dialogue(self):
        """Positive test for the 'update' method: the input message is for an existing dialogue."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        dialogue_reference = (
            msg.dialogue_reference[0],
            self.opponent_dialogues._generate_dialogue_nonce(),
        )
        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=dialogue_reference,
            message_id=msg.message_id + 1,
            target=msg.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        assert len(self.own_dialogues.dialogues) == 1

        dialogue = self.own_dialogues.update(valid_message_2_by_other)

        assert len(self.own_dialogues.dialogues) == 1
        assert dialogue is not None
        assert dialogue.last_message.dialogue_reference == dialogue_reference
        assert dialogue.last_message.message_id == valid_message_2_by_other.message_id
        assert dialogue.last_message.target == valid_message_2_by_other.target
        assert dialogue.last_message.performative == DefaultMessage.Performative.BYTES
        assert dialogue.last_message.content == b"Hello back"

    def test_update_positive_existing_dialogue_2(self):
        """Positive test for the 'update' method: the input message is for an existing dialogue from the original sender."""
        msg_1, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        opponent_dialogue_1 = self.opponent_dialogues.update(msg_1)

        msg_2 = dialogue.reply(
            performative=DefaultMessage.Performative.BYTES, content=b"Hello again",
        )

        opponent_dialogue_2 = self.opponent_dialogues.update(msg_2)

        assert opponent_dialogue_1 == opponent_dialogue_2

    def test_update_negative_invalid_label(self):
        """Negative test for the 'update' method: dialogue is not extendable with the input message."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=0,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_other.sender = self.opponent_address
        invalid_message_1_by_other.to = self.agent_address

        assert not self.own_dialogues.update(invalid_message_1_by_other)

    def test_update_negative_new_dialogue_by_self(self):
        """Negative test for the 'update' method: the message is not by the counterparty."""
        assert len(self.own_dialogues.dialogues) == 0

        with pytest.raises(AEAEnforceError) as cm:
            self.own_dialogues.update(self.valid_message_1_by_self)
        assert (
            str(cm.value)
            == "Invalid 'update' usage. Update must only be used with a message by another agent."
        )

        assert len(self.own_dialogues.dialogues) == 0

    def test_update_negative_no_to(self):
        """Negative test for the 'update' method: the 'to' field of the input message is not set."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_other.sender = self.opponent_address

        assert len(self.own_dialogues.dialogues) == 0

        with pytest.raises(AEAEnforceError) as cm:
            self.own_dialogues.update(invalid_message_1_by_other)
        assert str(cm.value) == "The message's 'to' field is not set {}".format(
            invalid_message_1_by_other
        )

        assert len(self.own_dialogues.dialogues) == 0

    def test_update_negative_no_sender(self):
        """Negative test for the 'update' method: the 'sender' field of the input message is not set."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_other.to = self.agent_address

        assert len(self.own_dialogues.dialogues) == 0

        with pytest.raises(AEAEnforceError) as cm:
            self.own_dialogues.update(invalid_message_1_by_other)
        assert (
            str(cm.value)
            == "Invalid 'update' usage. Update must only be used with a message by another agent."
        )

        assert len(self.own_dialogues.dialogues) == 0

    def test_update_negative_no_matching_to(self):
        """Negative test for the 'update' method: the 'to' field of the input message does not match self address."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        invalid_message_1_by_other.to = self.agent_address + "wrong_stuff"
        invalid_message_1_by_other.sender = self.opponent_address

        assert len(self.own_dialogues.dialogues) == 0

        with pytest.raises(AEAEnforceError) as cm:
            self.own_dialogues.update(invalid_message_1_by_other)
        assert (
            str(cm.value)
            == "Message to and dialogue self address do not match. Got 'to=agent 1wrong_stuff' expected 'to=agent 1'."
        )

        assert len(self.own_dialogues.dialogues) == 0

    def test_update_negative_invalid_message(self):
        """Negative test for the 'update' method: the message is invalid."""
        invalid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.ERROR,
            error_code=DefaultMessage.ErrorCode.INVALID_MESSAGE,
            error_msg="some_error_message",
            error_data={"some_data": b"some_bytes"},
        )
        invalid_message_1_by_other.sender = self.opponent_address
        invalid_message_1_by_other.to = self.agent_address

        assert len(self.own_dialogues.dialogues) == 0

        dialogue = self.own_dialogues.update(invalid_message_1_by_other)

        assert dialogue is None

        assert len(self.own_dialogues.dialogues) == 0

    def test_update_negative_existing_dialogue_non_nonexistent(self):
        """Negative test for the 'update' method: the dialogue referred by the input message does not exist."""
        _, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        invalid_message_2_by_other = DefaultMessage(
            dialogue_reference=(str(2), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_2_by_other.sender = self.opponent_address
        invalid_message_2_by_other.to = self.agent_address

        updated_dialogue = self.own_dialogues.update(invalid_message_2_by_other)

        assert updated_dialogue is None
        last_message = self.own_dialogues.dialogues[
            dialogue.dialogue_label
        ].last_message
        assert (
            last_message.dialogue_reference[0] != ""
            and last_message.dialogue_reference[1] == ""
        )
        assert (
            self.own_dialogues.dialogues[
                dialogue.dialogue_label
            ].last_message.message_id
            == 1
        )
        assert (
            self.own_dialogues.dialogues[dialogue.dialogue_label].last_message.target
            == 0
        )
        assert (
            self.own_dialogues.dialogues[
                dialogue.dialogue_label
            ].last_message.performative
            == DefaultMessage.Performative.BYTES
        )
        assert (
            self.own_dialogues.dialogues[dialogue.dialogue_label].last_message.content
            == b"Hello"
        )

    def test_complete_dialogue_reference_positive(self,):
        """Positive test for the '_complete_dialogue_reference' method."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(
                msg.dialogue_reference[0],
                self.opponent_dialogues._generate_dialogue_nonce(),
            ),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        self.own_dialogues._complete_dialogue_reference(valid_message_2_by_other)

        assert (
            self.own_dialogues.dialogues[
                dialogue.dialogue_label
            ].dialogue_label.dialogue_reference
            == valid_message_2_by_other.dialogue_reference
        )

    def test_complete_dialogue_reference_negative_incorrect_reference(self,):
        """Negative test for the '_complete_dialogue_reference' method: the input message has invalid dialogue reference."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        wrong_own_ref = (
            "wrong reference"  # if correct, would be  msg.dialogue_reference[0]
        )
        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(
                wrong_own_ref,
                self.opponent_dialogues._generate_dialogue_nonce(),
            ),
            message_id=msg.message_id + 1,
            target=msg.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        self.own_dialogues._complete_dialogue_reference(valid_message_2_by_other)
        assert (
            self.own_dialogues.dialogues[
                dialogue.dialogue_label
            ].dialogue_label.dialogue_reference
            == msg.dialogue_reference
        )

    def test_get_dialogue_positive_1(self):
        """Positive test for the 'get_dialogue' method: the dialogue is self initiated and the second message is by the other agent."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(
                msg.dialogue_reference[0],
                self.opponent_dialogues._generate_dialogue_nonce(),
            ),
            message_id=msg.message_id + 1,
            target=msg.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        self.own_dialogues._complete_dialogue_reference(valid_message_2_by_other)

        assert (
            self.own_dialogues.dialogues[
                dialogue.dialogue_label
            ].dialogue_label.dialogue_reference
            == valid_message_2_by_other.dialogue_reference
        )

        retrieved_dialogue = self.own_dialogues.get_dialogue(valid_message_2_by_other)

        assert retrieved_dialogue.dialogue_label == dialogue.dialogue_label

    def test_get_dialogue_positive_2(self):
        """Positive test for the 'get_dialogue' method: the dialogue is other initiated and the second message is by this agent."""
        valid_message_1_by_other = DefaultMessage(
            dialogue_reference=(str(1), ""),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        valid_message_1_by_other.sender = self.opponent_address
        valid_message_1_by_other.to = self.agent_address

        dialogue = self.own_dialogues.update(valid_message_1_by_other)

        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=dialogue.dialogue_label.dialogue_reference,
            message_id=valid_message_1_by_other.message_id + 1,
            target=valid_message_1_by_other.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.agent_address
        valid_message_2_by_other.to = self.opponent_address

        retrieved_dialogue = self.own_dialogues.get_dialogue(valid_message_2_by_other)

        assert retrieved_dialogue is not None
        assert retrieved_dialogue.dialogue_label == dialogue.dialogue_label

    def test_get_dialogue_negative_invalid_reference(self):
        """Negative test for the 'get_dialogue' method: the input message has invalid dialogue reference."""
        msg, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        valid_message_2_by_other = DefaultMessage(
            dialogue_reference=(
                msg.dialogue_reference[0],
                self.opponent_dialogues._generate_dialogue_nonce(),
            ),
            message_id=msg.message_id + 1,
            target=msg.message_id,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        valid_message_2_by_other.sender = self.opponent_address
        valid_message_2_by_other.to = self.agent_address

        dialogue = self.own_dialogues.update(valid_message_2_by_other)
        assert dialogue is not None

        invalid_message_3_by_self = DefaultMessage(
            dialogue_reference=(str(2), str(1)),
            message_id=2,
            target=1,
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello back",
        )
        invalid_message_3_by_self.sender = self.agent_address
        invalid_message_3_by_self.to = self.opponent_address

        retrieved_dialogue = self.own_dialogues.get_dialogue(invalid_message_3_by_self)

        assert retrieved_dialogue is None

    def test_get_latest_label(self):
        """Positive test for the 'get_latest_label' method."""
        pass

    def test_get_dialogue_from_label_positive(self):
        """Positive test for the 'get_dialogue_from_label' method."""
        _, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        retrieved_dialogue = self.own_dialogues._get_dialogue_from_label(
            dialogue.dialogue_label
        )
        assert retrieved_dialogue.dialogue_label == dialogue.dialogue_label

    def test_get_dialogue_from_label_negative_incorrect_input_label(self):
        """Negative test for the 'get_dialogue_from_label' method: the input dialogue label does not exist."""
        _, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )

        incorrect_label = DialogueLabel(
            (str(1), "error"), self.opponent_address, self.agent_address
        )

        retrieved_dialogue = self.own_dialogues._get_dialogue_from_label(
            incorrect_label
        )
        assert retrieved_dialogue is None

    def test_create_self_initiated_positive(self):
        """Positive test for the '_create_self_initiated' method."""
        assert len(self.own_dialogues.dialogues) == 0

        self.own_dialogues._create_self_initiated(
            self.opponent_address, (str(1), ""), Dialogue.Role.ROLE1
        )
        assert len(self.own_dialogues.dialogues) == 1

    def test_create_self_initiated_negative_invalid_dialogue_reference(self):
        """Negative test for the '_create_self_initiated' method: invalid dialogue reference"""
        pass

    def test_create_opponent_initiated_positive(self):
        """Positive test for the '_create_opponent_initiated' method."""
        assert len(self.own_dialogues.dialogues) == 0

        self.own_dialogues._create_opponent_initiated(
            self.opponent_address, (str(1), ""), Dialogue.Role.ROLE2
        )
        assert len(self.own_dialogues.dialogues) == 1

    def test_create_opponent_initiated_negative_invalid_input_dialogue_reference(self):
        """Negative test for the '_create_opponent_initiated' method: input dialogue label has invalid dialogue reference."""
        assert len(self.own_dialogues.dialogues) == 0

        try:
            self.own_dialogues._create_opponent_initiated(
                self.opponent_address, ("", str(1)), Dialogue.Role.ROLE2
            )
            result = True
        except AEAEnforceError:
            result = False

        assert not result
        assert len(self.own_dialogues.dialogues) == 0

    def test_create_with_message(self):
        """Positive test for create with message."""
        msg = DefaultMessage(
            dialogue_reference=self.own_dialogues.new_self_initiated_dialogue_reference(),
            performative=DefaultMessage.Performative.BYTES,
            content=b"Hello",
        )
        self.own_dialogues.create_with_message("opponent", msg)

    def test__create_positive(self):
        """Positive test for the '_create' method."""
        pass

    def test__create_negative_incomplete_dialogue_label_present(self):
        """Negative test for the '_create' method: incomplete dialogue label already present."""
        pass

    def test__create_negative_dialogue_label_present(self):
        """Negative test for the '_create' method: dialogue label already present."""
        pass

    def test_generate_dialogue_nonce(self):
        """Test the '_generate_dialogue_nonce' method."""
        nonce = self.own_dialogues._generate_dialogue_nonce()
        assert (
            isinstance(nonce, str)
            and nonce != ""
            and len(nonce) == DialogueLabel.NONCE_BYTES_NB * 2
        )
        second_nonce = self.own_dialogues._generate_dialogue_nonce()
        assert nonce != second_nonce

    def test_get_dialogues_with_counterparty(self):
        """Test get dialogues with counterparty."""
        assert (
            self.own_dialogues.get_dialogues_with_counterparty(self.opponent_address)
            == []
        )
        _, dialogue = self.own_dialogues.create(
            self.opponent_address, DefaultMessage.Performative.BYTES, content=b"Hello"
        )
        assert self.own_dialogues.get_dialogues_with_counterparty(
            self.opponent_address
        ) == [dialogue]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the decision maker class."""

import copy
import logging
from enum import Enum
from typing import Any, Dict, List, Optional, cast

from aea.common import Address
from aea.crypto.wallet import Wallet
from aea.decision_maker.base import DecisionMakerHandler as BaseDecisionMakerHandler
from aea.decision_maker.base import OwnershipState as BaseOwnershipState
from aea.decision_maker.base import Preferences as BasePreferences
from aea.exceptions import enforce
from aea.helpers.preference_representations.base import (
    linear_utility,
    logarithmic_utility,
)
from aea.helpers.transaction.base import SignedMessage, SignedTransaction, Terms
from aea.identity.base import Identity
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.signing.dialogues import SigningDialogue
from aea.protocols.signing.dialogues import SigningDialogues as BaseSigningDialogues
from aea.protocols.signing.message import SigningMessage
from aea.protocols.state_update.dialogues import StateUpdateDialogue
from aea.protocols.state_update.dialogues import (
    StateUpdateDialogues as BaseStateUpdateDialogues,
)
from aea.protocols.state_update.message import StateUpdateMessage


CurrencyHoldings = Dict[str, int]  # a map from identifier to quantity
GoodHoldings = Dict[str, int]  # a map from identifier to quantity
UtilityParams = Dict[str, float]  # a map from identifier to quantity
ExchangeParams = Dict[str, float]  # a map from identifier to quantity

QUANTITY_SHIFT = 100

_default_logger = logging.getLogger(__name__)


class SigningDialogues(BaseSigningDialogues):
    """This class keeps track of all oef_search dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom dialogues are maintained
        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return SigningDialogue.Role.DECISION_MAKER

        BaseSigningDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            **kwargs,
        )


class StateUpdateDialogues(BaseStateUpdateDialogues):
    """This class keeps track of all oef_search dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom dialogues are maintained
        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return StateUpdateDialogue.Role.DECISION_MAKER

        BaseStateUpdateDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            **kwargs,
        )


class GoalPursuitReadiness:
    """The goal pursuit readiness."""

    class Status(Enum):
        """
        The enum of the readiness status.

        In particular, it can be one of the following:

        - Status.READY: when the agent is ready to pursuit its goal
        - Status.NOT_READY: when the agent is not ready to pursuit its goal
        """

        READY = "ready"
        NOT_READY = "not_ready"

    def __init__(self):
        """Instantiate the goal pursuit readiness."""
        self._status = GoalPursuitReadiness.Status.NOT_READY

    @property
    def is_ready(self) -> bool:
        """Get the readiness."""
        return self._status.value == GoalPursuitReadiness.Status.READY.value

    def update(self, new_status: Status) -> None:
        """
        Update the goal pursuit readiness.

        :param new_status: the new status
        :return: None
        """
        self._status = new_status


class OwnershipState(BaseOwnershipState):
    """Represent the ownership state of an agent (can proxy a ledger)."""

    def __init__(self):
        """
        Instantiate an ownership state object.

        :param decision_maker: the decision maker
        """
        self._amount_by_currency_id = None  # type: Optional[CurrencyHoldings]
        self._quantities_by_good_id = None  # type: Optional[GoodHoldings]

    def set(  # pylint: disable=arguments-differ
        self,
        amount_by_currency_id: CurrencyHoldings = None,
        quantities_by_good_id: GoodHoldings = None,
        **kwargs,
    ) -> None:
        """
        Set values on the ownership state.

        :param amount_by_currency_id: the currency endowment of the agent in this state.
        :param quantities_by_good_id: the good endowment of the agent in this state.
        """
        if amount_by_currency_id is None:  # pragma: nocover
            raise ValueError("Must provide amount_by_currency_id.")
        if quantities_by_good_id is None:  # pragma: nocover
            raise ValueError("Must provide quantities_by_good_id.")
        enforce(
            not self.is_initialized,
            "Cannot apply state update, current state is already initialized!",
        )

        self._amount_by_currency_id = copy.copy(amount_by_currency_id)
        self._quantities_by_good_id = copy.copy(quantities_by_good_id)

    def apply_delta(  # pylint: disable=arguments-differ
        self,
        delta_amount_by_currency_id: Dict[str, int] = None,
        delta_quantities_by_good_id: Dict[str, int] = None,
        **kwargs,
    ) -> None:
        """
        Apply a state update to the ownership state.

        This method is used to apply a raw state update without a transaction.

        :param delta_amount_by_currency_id: the delta in the currency amounts
        :param delta_quantities_by_good_id: the delta in the quantities by good
        :return: None
        """
        if delta_amount_by_currency_id is None:  # pragma: nocover
            raise ValueError("Must provide delta_amount_by_currency_id.")
        if delta_quantities_by_good_id is None:  # pragma: nocover
            raise ValueError("Must provide delta_quantities_by_good_id.")
        if self._amount_by_currency_id is None or self._quantities_by_good_id is None:
            raise ValueError(  # pragma: nocover
                "Cannot apply state update, current state is not initialized!"
            )
        enforce(
            all(
                [
                    key in self._amount_by_currency_id
                    for key in delta_amount_by_currency_id.keys()
                ]
            ),
            "Invalid keys present in delta_amount_by_currency_id.",
        )
        enforce(
            all(
                [
                    key in self._quantities_by_good_id
                    for key in delta_quantities_by_good_id.keys()
                ]
            ),
            "Invalid keys present in delta_quantities_by_good_id.",
        )

        for currency_id, amount_delta in delta_amount_by_currency_id.items():
            self._amount_by_currency_id[currency_id] += amount_delta

        for good_id, quantity_delta in delta_quantities_by_good_id.items():
            self._quantities_by_good_id[good_id] += quantity_delta

    @property
    def is_initialized(self) -> bool:
        """Get the initialization status."""
        return (
            self._amount_by_currency_id is not None
            and self._quantities_by_good_id is not None
        )

    @property
    def amount_by_currency_id(self) -> CurrencyHoldings:
        """Get currency holdings in this state."""
        if self._amount_by_currency_id is None:
            raise ValueError("amount_by_currency_id is not set!")
        return copy.copy(self._amount_by_currency_id)

    @property
    def quantities_by_good_id(self) -> GoodHoldings:
        """Get good holdings in this state."""
        if self._quantities_by_good_id is None:
            raise ValueError("quantities_by_good_id is not set!")
        return copy.copy(self._quantities_by_good_id)

    def is_affordable_transaction(self, terms: Terms) -> bool:
        """
        Check if the transaction is affordable (and consistent).

        E.g. check that the agent state has enough money if it is a buyer or enough holdings if it is a seller.
        Note, the agent is the sender of the transaction message by design.

        :param terms: the transaction terms
        :return: True if the transaction is legal wrt the current state, false otherwise.
        """
        if all(amount == 0 for amount in terms.amount_by_currency_id.values()) and all(
            quantity == 0 for quantity in terms.quantities_by_good_id.values()
        ):
            # reject the transaction when there is no wealth exchange
            result = False
        elif all(
            amount <= 0 for amount in terms.amount_by_currency_id.values()
        ) and all(quantity >= 0 for quantity in terms.quantities_by_good_id.values()):
            # check if the agent has the money to cover the sender_amount (the agent=sender is the buyer)
            result = all(
                self.amount_by_currency_id[currency_id] >= -amount
                for currency_id, amount in terms.amount_by_currency_id.items()
            )
        elif all(
            amount >= 0 for amount in terms.amount_by_currency_id.values()
        ) and all(quantity <= 0 for quantity in terms.quantities_by_good_id.values()):
            # check if the agent has the goods (the agent=sender is the seller).
            result = all(
                self.quantities_by_good_id[good_id] >= -quantity
                for good_id, quantity in terms.quantities_by_good_id.items()
            )
        else:
            result = False
        return result

    def is_affordable(self, terms: Terms) -> bool:
        """
        Check if the tx is affordable.

        :param terms: the transaction terms
        :return: whether the transaction is affordable or not
        """
        if self.is_initialized:
            is_affordable = self.is_affordable_transaction(terms)
        else:
            _default_logger.warning(
                "Cannot verify whether transaction is affordable as ownership state is not initialized. Assuming it is!"
            )
            is_affordable = True
        return is_affordable

    def update(self, terms: Terms) -> None:
        """
        Update the agent state from a transaction.

        :param terms: the transaction terms
        :return: None
        """
        if self._amount_by_currency_id is None or self._quantities_by_good_id is None:
            raise ValueError(  # pragma: nocover
                "Cannot apply state update, current state is not initialized!"
            )
        for currency_id, amount_delta in terms.amount_by_currency_id.items():
            self._amount_by_currency_id[currency_id] += amount_delta

        for good_id, quantity_delta in terms.quantities_by_good_id.items():
            self._quantities_by_good_id[good_id] += quantity_delta

    def apply_transactions(self, list_of_terms: List[Terms]) -> "OwnershipState":
        """
        Apply a list of transactions to (a copy of) the current state.

        :param list_of_terms: the sequence of transaction terms.
        :return: the final state.
        """
        new_state = copy.copy(self)
        for terms in list_of_terms:
            new_state.update(terms)

        return new_state

    def __copy__(self) -> "OwnershipState":
        """Copy the object."""
        state = OwnershipState()
        if self.is_initialized:
            state._amount_by_currency_id = self.amount_by_currency_id
            state._quantities_by_good_id = self.quantities_by_good_id
        return state


class Preferences(BasePreferences):
    """Class to represent the preferences."""

    def __init__(self):
        """Instantiate an agent preference object."""
        self._exchange_params_by_currency_id = None  # type: Optional[ExchangeParams]
        self._utility_params_by_good_id = None  # type: Optional[UtilityParams]
        self._quantity_shift = QUANTITY_SHIFT

    def set(  # pylint: disable=arguments-differ
        self,
        exchange_params_by_currency_id: ExchangeParams = None,
        utility_params_by_good_id: UtilityParams = None,
        **kwargs,
    ) -> None:
        """
        Set values on the preferences.

        :param exchange_params_by_currency_id: the exchange params.
        :param utility_params_by_good_id: the utility params for every asset.
        """
        if exchange_params_by_currency_id is None:  # pragma: nocover
            raise ValueError("Must provide exchange_params_by_currency_id.")
        if utility_params_by_good_id is None:  # pragma: nocover
            raise ValueError("Must provide utility_params_by_good_id.")
        enforce(
            not self.is_initialized,
            "Cannot apply preferences update, preferences already initialized!",
        )

        self._exchange_params_by_currency_id = copy.copy(exchange_params_by_currency_id)
        self._utility_params_by_good_id = copy.copy(utility_params_by_good_id)

    @property
    def is_initialized(self) -> bool:
        """
        Get the initialization status.

        Returns True if exchange_params_by_currency_id and utility_params_by_good_id are not None.
        """
        return (self._exchange_params_by_currency_id is not None) and (
            self._utility_params_by_good_id is not None
        )

    @property
    def exchange_params_by_currency_id(self) -> ExchangeParams:
        """Get exchange parameter for each currency."""
        if self._exchange_params_by_currency_id is None:
            raise ValueError("ExchangeParams not set!")
        return self._exchange_params_by_currency_id

    @property
    def utility_params_by_good_id(self) -> UtilityParams:
        """Get utility parameter for each good."""
        if self._utility_params_by_good_id is None:
            raise ValueError("UtilityParams not set!")
        return self._utility_params_by_good_id

    def logarithmic_utility(self, quantities_by_good_id: GoodHoldings) -> float:
        """
        Compute agent's utility given her utility function params and a good bundle.

        :param quantities_by_good_id: the good holdings (dictionary) with the identifier (key) and quantity (value) for each good
        :return: utility value
        """
        enforce(self.is_initialized, "Preferences params not set!")
        result = logarithmic_utility(
            self.utility_params_by_good_id, quantities_by_good_id, self._quantity_shift
        )
        return result

    def linear_utility(self, amount_by_currency_id: CurrencyHoldings) -> float:
        """
        Compute agent's utility given her utility function params and a currency bundle.

        :param amount_by_currency_id: the currency holdings (dictionary) with the identifier (key) and quantity (value) for each currency
        :return: utility value
        """
        enforce(self.is_initialized, "Preferences params not set!")
        result = linear_utility(
            self.exchange_params_by_currency_id, amount_by_currency_id
        )
        return result

    def utility(
        self,
        quantities_by_good_id: GoodHoldings,
        amount_by_currency_id: CurrencyHoldings,
    ) -> float:
        """
        Compute the utility given the good and currency holdings.

        :param quantities_by_good_id: the good holdings
        :param amount_by_currency_id: the currency holdings
        :return: the utility value.
        """
        enforce(self.is_initialized, "Preferences params not set!")
        goods_score = self.logarithmic_utility(quantities_by_good_id)
        currency_score = self.linear_utility(amount_by_currency_id)
        score = goods_score + currency_score
        return score

    def marginal_utility(  # pylint: disable=arguments-differ
        self,
        ownership_state: BaseOwnershipState,
        delta_quantities_by_good_id: Optional[GoodHoldings] = None,
        delta_amount_by_currency_id: Optional[CurrencyHoldings] = None,
        **kwargs,
    ) -> float:
        """
        Compute the marginal utility.

        :param ownership_state: the ownership state against which to compute the marginal utility.
        :param delta_quantities_by_good_id: the change in good holdings
        :param delta_amount_by_currency_id: the change in money holdings
        :return: the marginal utility score
        """
        enforce(self.is_initialized, "Preferences params not set!")
        ownership_state = cast(OwnershipState, ownership_state)
        current_goods_score = self.logarithmic_utility(
            ownership_state.quantities_by_good_id
        )
        current_currency_score = self.linear_utility(
            ownership_state.amount_by_currency_id
        )
        new_goods_score = current_goods_score
        new_currency_score = current_currency_score
        if delta_quantities_by_good_id is not None:
            new_quantities_by_good_id = {
                good_id: quantity + delta_quantities_by_good_id[good_id]
                for good_id, quantity in ownership_state.quantities_by_good_id.items()
            }
            new_goods_score = self.logarithmic_utility(new_quantities_by_good_id)
        if delta_amount_by_currency_id is not None:
            new_amount_by_currency_id = {
                currency: amount + delta_amount_by_currency_id[currency]
                for currency, amount in ownership_state.amount_by_currency_id.items()
            }
            new_currency_score = self.linear_utility(new_amount_by_currency_id)
        marginal_utility = (
            new_goods_score
            + new_currency_score
            - current_goods_score
            - current_currency_score
        )
        return marginal_utility

    def utility_diff_from_transaction(
        self, ownership_state: BaseOwnershipState, terms: Terms
    ) -> float:
        """
        Simulate a transaction and get the resulting utility difference (taking into account the fee).

        :param ownership_state: the ownership state against which to apply the transaction.
        :param terms: the transaction terms.
        :return: the score.
        """
        enforce(self.is_initialized, "Preferences params not set!")
        ownership_state = cast(OwnershipState, ownership_state)
        current_score = self.utility(
            quantities_by_good_id=ownership_state.quantities_by_good_id,
            amount_by_currency_id=ownership_state.amount_by_currency_id,
        )
        new_ownership_state = ownership_state.apply_transactions([terms])
        new_score = self.utility(
            quantities_by_good_id=new_ownership_state.quantities_by_good_id,
            amount_by_currency_id=new_ownership_state.amount_by_currency_id,
        )
        score_difference = new_score - current_score
        return score_difference

    def is_utility_enhancing(
        self, ownership_state: BaseOwnershipState, terms: Terms
    ) -> bool:
        """
        Check if the tx is utility enhancing.

        :param ownership_state: the ownership state against which to apply the transaction.
        :param terms: the transaction terms
        :return: whether the transaction is utility enhancing or not
        """
        if self.is_initialized and ownership_state.is_initialized:
            is_utility_enhancing = (
                self.utility_diff_from_transaction(ownership_state, terms) >= 0.0
            )
        else:
            _default_logger.warning(
                "Cannot verify whether transaction improves utility as preferences are not initialized. Assuming it does!"
            )
            is_utility_enhancing = True
        return is_utility_enhancing

    def __copy__(self) -> "Preferences":
        """Copy the object."""
        preferences = Preferences()
        if self.is_initialized:
            preferences._exchange_params_by_currency_id = (
                self.exchange_params_by_currency_id
            )
            preferences._utility_params_by_good_id = self.utility_params_by_good_id
        return preferences


class DecisionMakerHandler(BaseDecisionMakerHandler):
    """This class implements the decision maker."""

    def __init__(self, identity: Identity, wallet: Wallet):
        """
        Initialize the decision maker.

        :param identity: the identity
        :param wallet: the wallet
        """
        kwargs = {
            "goal_pursuit_readiness": GoalPursuitReadiness(),
            "ownership_state": OwnershipState(),
            "preferences": Preferences(),
        }
        super().__init__(
            identity=identity, wallet=wallet, **kwargs,
        )
        self.signing_dialogues = SigningDialogues(self.self_address)
        self.state_update_dialogues = StateUpdateDialogues(self.self_address)

    def handle(self, message: Message) -> None:
        """
        Handle an internal message from the skills.

        :param message: the internal message
        :return: None
        """
        if isinstance(message, SigningMessage):
            self._handle_signing_message(message)
        elif isinstance(message, StateUpdateMessage):
            self._handle_state_update_message(message)
        else:  # pragma: no cover
            self.logger.error(
                "[{}]: cannot handle message={} of type={}".format(
                    self.agent_name, message, type(message)
                )
            )

    def _handle_signing_message(self, signing_msg: SigningMessage) -> None:
        """
        Handle a signing message.

        :param signing_msg: the transaction message
        :return: None
        """
        if not self.context.goal_pursuit_readiness.is_ready:
            self.logger.debug(
                "[{}]: Preferences and ownership state not initialized!".format(
                    self.agent_name
                )
            )

        signing_dialogue = cast(
            Optional[SigningDialogue], self.signing_dialogues.update(signing_msg)
        )
        if signing_dialogue is None:  # pragma: no cover
            self.logger.error(
                "[{}]: Could not construct signing dialogue. Aborting!".format(
                    self.agent_name
                )
            )
            return

        # check if the transaction is acceptable and process it accordingly
        if signing_msg.performative == SigningMessage.Performative.SIGN_MESSAGE:
            self._handle_message_signing(signing_msg, signing_dialogue)
        elif signing_msg.performative == SigningMessage.Performative.SIGN_TRANSACTION:
            self._handle_transaction_signing(signing_msg, signing_dialogue)
        else:  # pragma: no cover
            self.logger.error(
                "[{}]: Unexpected transaction message performative".format(
                    self.agent_name
                )
            )

    def _handle_message_signing(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle a message for signing.

        :param signing_msg: the signing message
        :param signing_dialogue: the signing dialogue
        :return: None
        """
        performative = SigningMessage.Performative.ERROR
        kwargs = {
            "error_code": SigningMessage.ErrorCode.UNSUCCESSFUL_MESSAGE_SIGNING,
        }  # type: Dict[str, Any]
        if self._is_acceptable_for_signing(signing_msg):
            signed_message = self.wallet.sign_message(
                signing_msg.raw_message.ledger_id,
                signing_msg.raw_message.body,
                signing_msg.raw_message.is_deprecated_mode,
            )
            if signed_message is not None:
                performative = SigningMessage.Performative.SIGNED_MESSAGE
                kwargs.pop("error_code")
                kwargs["signed_message"] = SignedMessage(
                    signing_msg.raw_message.ledger_id,
                    signed_message,
                    signing_msg.raw_message.is_deprecated_mode,
                )
        signing_msg_response = signing_dialogue.reply(
            performative=performative, target_message=signing_msg, **kwargs,
        )
        self.message_out_queue.put(signing_msg_response)

    def _handle_transaction_signing(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle a transaction for signing.

        :param signing_msg: the signing message
        :param signing_dialogue: the signing dialogue
        :return: None
        """
        performative = SigningMessage.Performative.ERROR
        kwargs = {
            "error_code": SigningMessage.ErrorCode.UNSUCCESSFUL_TRANSACTION_SIGNING,
        }  # type: Dict[str, Any]
        if self._is_acceptable_for_signing(signing_msg):
            signed_tx = self.wallet.sign_transaction(
                signing_msg.raw_transaction.ledger_id, signing_msg.raw_transaction.body
            )
            if signed_tx is not None:
                performative = SigningMessage.Performative.SIGNED_TRANSACTION
                kwargs.pop("error_code")
                kwargs["signed_transaction"] = SignedTransaction(
                    signing_msg.raw_transaction.ledger_id, signed_tx
                )
        signing_msg_response = signing_dialogue.reply(
            performative=performative, target_message=signing_msg, **kwargs,
        )
        self.message_out_queue.put(signing_msg_response)

    def _is_acceptable_for_signing(self, signing_msg: SigningMessage) -> bool:
        """
        Check if the tx message is acceptable for signing.

        :param signing_msg: the transaction message
        :return: whether the transaction is acceptable or not
        """
        result = self.context.preferences.is_utility_enhancing(
            self.context.ownership_state, signing_msg.terms
        ) and self.context.ownership_state.is_affordable(signing_msg.terms)
        return result

    def _handle_state_update_message(
        self, state_update_msg: StateUpdateMessage
    ) -> None:
        """
        Handle a state update message.

        :param state_update_message: the state update message
        :return: None
        """
        state_update_dialogue = cast(
            Optional[StateUpdateDialogue],
            self.state_update_dialogues.update(state_update_msg),
        )
        if state_update_dialogue is None:  # pragma: no cover
            self.logger.error(
                "[{}]: Could not construct state_update dialogue. Aborting!".format(
                    self.agent_name
                )
            )
            return

        if state_update_msg.performative == StateUpdateMessage.Performative.INITIALIZE:
            self.logger.warning(
                "[{}]: Applying ownership_state and preferences initialization!".format(
                    self.agent_name
                )
            )
            self.context.ownership_state.set(
                amount_by_currency_id=state_update_msg.amount_by_currency_id,
                quantities_by_good_id=state_update_msg.quantities_by_good_id,
            )
            self.context.preferences.set(
                exchange_params_by_currency_id=state_update_msg.exchange_params_by_currency_id,
                utility_params_by_good_id=state_update_msg.utility_params_by_good_id,
            )
            self.context.goal_pursuit_readiness.update(
                GoalPursuitReadiness.Status.READY
            )
        elif state_update_msg.performative == StateUpdateMessage.Performative.APPLY:
            self.logger.info("[{}]: Applying state update!".format(self.agent_name))
            self.context.ownership_state.apply_delta(
                delta_amount_by_currency_id=state_update_msg.amount_by_currency_id,
                delta_quantities_by_good_id=state_update_msg.quantities_by_good_id,
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the code-blocks in the standalone-transaction.md file."""

import logging
import os
from unittest.mock import patch

import pytest

from aea.test_tools.test_cases import BaseAEATestCase

from tests.conftest import CUR_PATH, MAX_FLAKY_RERUNS_INTEGRATION, ROOT_DIR
from tests.test_docs.helper import extract_code_blocks, extract_python_code
from tests.test_docs.test_standalone_transaction.standalone_transaction import (
    logger,
    run,
)


MD_FILE = "docs/standalone-transaction.md"
PY_FILE = "test_docs/test_standalone_transaction/standalone_transaction.py"

test_logger = logging.getLogger(__name__)


class TestStandaloneTransaction(BaseAEATestCase):
    """This class contains the tests for the code-blocks in the agent-vs-aea.md file."""

    @classmethod
    def _patch_logger(cls):

        cls.patch_logger_info = patch.object(logger, "info")
        cls.mocked_logger_info = cls.patch_logger_info.__enter__()

    @classmethod
    def _unpatch_logger(cls):
        cls.mocked_logger_info.__exit__()

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        BaseAEATestCase.setup_class()
        cls._patch_logger()
        doc_path = os.path.join(ROOT_DIR, MD_FILE)
        cls.code_blocks = extract_code_blocks(filepath=doc_path, filter="python")
        test_code_path = os.path.join(CUR_PATH, PY_FILE)
        cls.python_file = extract_python_code(test_code_path)

    def test_read_md_file(self):
        """Test the last code block, that is the full listing of the demo from the Markdown."""
        assert (
            self.code_blocks[-1] == self.python_file
        ), "Files must be exactly the same."

    @pytest.mark.integration(reruns=MAX_FLAKY_RERUNS_INTEGRATION)
    def test_run_end_to_end(self):
        """Run the transaction from the file."""
        try:
            run()
            self.mocked_logger_info.assert_any_call("Transaction complete.")
        except RuntimeError:
            test_logger.info("RuntimeError: Some transactions have failed")

    def test_code_blocks_exist(self):
        """Test that all the code-blocks exist in the python file."""
        for blocks in self.code_blocks:
            assert (
                blocks in self.python_file
            ), "Code-block doesn't exist in the python file."



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the crypto/helpers module."""

import logging
import os
from pathlib import Path
from unittest.mock import mock_open, patch

import pytest
import requests

from aea.crypto.cosmos import CosmosCrypto
from aea.crypto.ethereum import EthereumCrypto
from aea.crypto.fetchai import FetchAICrypto
from aea.crypto.helpers import (
    create_private_key,
    try_generate_testnet_wealth,
    try_validate_private_key_path,
    verify_or_create_private_keys,
)

from tests.conftest import (
    COSMOS_PRIVATE_KEY_FILE,
    CUR_PATH,
    ETHEREUM_PRIVATE_KEY_FILE,
    ETHEREUM_PRIVATE_KEY_PATH,
    FETCHAI_PRIVATE_KEY_PATH,
)


logger = logging.getLogger(__name__)


class ResponseMock:
    """Mock for response class."""

    text = "some text"

    def __init__(self, status_code=200):
        """Initialise response mock."""
        self.status_code = status_code


class TestHelperFile:
    """Test helper module in aea/crypto."""

    def tests_private_keys(self):
        """Test the private keys."""
        try_validate_private_key_path(
            FetchAICrypto.identifier, FETCHAI_PRIVATE_KEY_PATH
        )
        with pytest.raises(SystemExit):
            private_key_path = os.path.join(
                CUR_PATH, "data", "fet_private_key_wrong.txt"
            )
            try_validate_private_key_path(FetchAICrypto.identifier, private_key_path)

        try_validate_private_key_path(
            EthereumCrypto.identifier, ETHEREUM_PRIVATE_KEY_PATH
        )
        with pytest.raises(SystemExit):
            private_key_path = os.path.join(
                CUR_PATH, "data", "fet_private_key_wrong.txt"
            )
            try_validate_private_key_path(EthereumCrypto.identifier, private_key_path)

    @patch("aea.crypto.ethereum._default_logger")
    def tests_generate_wealth_ethereum(self, mock_logging):
        """Test generate wealth for ethereum."""
        address = "my_address"
        result = ResponseMock(status_code=500)
        with patch.object(requests, "get", return_value=result):
            try_generate_testnet_wealth(
                identifier=EthereumCrypto.identifier, address=address
            )
            assert mock_logging.error.called

        result.status_code = 200
        with patch.object(requests, "get", return_value=result):
            try_generate_testnet_wealth(
                identifier=EthereumCrypto.identifier, address=address
            )

    @patch("aea.crypto.ethereum.requests.post", return_value=ResponseMock())
    @patch("aea.crypto.ethereum.json.loads", return_value={"error_message": ""})
    def test_try_generate_testnet_wealth_error_resp_ethereum(self, *mocks):
        """Test try_generate_testnet_wealth error_resp."""
        try_generate_testnet_wealth(EthereumCrypto.identifier, "address")

    def test_try_validate_private_key_path_positive(self):
        """Test _validate_private_key_path positive result."""
        try_validate_private_key_path(
            FetchAICrypto.identifier, FETCHAI_PRIVATE_KEY_PATH
        )
        try_validate_private_key_path(
            EthereumCrypto.identifier, ETHEREUM_PRIVATE_KEY_PATH
        )

    @patch("builtins.open", mock_open())
    def test__create_ethereum_private_key_positive(self, *mocks):
        """Test _create_ethereum_private_key positive result."""
        create_private_key(EthereumCrypto.identifier, ETHEREUM_PRIVATE_KEY_FILE)

    @patch("builtins.open", mock_open())
    def test__create_cosmos_private_key_positive(self, *mocks):
        """Test _create_cosmos_private_key positive result."""
        create_private_key(CosmosCrypto.identifier, COSMOS_PRIVATE_KEY_FILE)

    @patch("aea.crypto.helpers.create_private_key")
    @patch("aea.crypto.helpers.try_validate_private_key_path")
    def test_verify_or_create_private_keys(self, *mocks):
        """Test _create_ethereum_private_key positive result."""
        verify_or_create_private_keys(Path(os.path.join(CUR_PATH, "data", "dummy_aea")))
